# pylint: disable = undefined-variable
DharmaConst.VARIANCE_MIN = 1
DharmaConst.VARIANCE_MAX = 8
DharmaConst.VARIABLE_MIN = 1
DharmaConst.VARIABLE_MAX = 4
DharmaConst.VARIANCE_TEMPLATE = "%s"
DharmaConst.MAX_REPEAT_POWER = 12
DharmaConst.LEAF_TRIGGER = 256
DharmaConst.URI_TABLE = {
    "images": "fuzzdata/samples/jpg/",
    "videos": "fuzzdata/samples/mp4/",
    "audios": "fuzzdata/samples/mp3/",
}
