__title__ = 'dharma'
__version__ = '1.3.2'
__license__ = 'MPL 2.0'
__description__ = 'Generation-based, context-free grammar fuzzer.'
__keywords__ = 'fuzzer fuzzing security testing generator'
__author__ = 'Christoph Diehl'
__author_email__ = 'cdiehl@mozilla.com'
__maintainer__ = 'Mozilla Fuzzing Team'
__maintainer_email__ = 'fuzzing@mozilla.com'
__url__ = 'https://github.com/MozillaSecurity/dharma'
__project_urls__ = {
    'Bug Tracker': 'https://github.com/MozillaSecurity/dharma/issues'
}
__download_url__ = "https://github.com/MozillaSecurity/dharma/tarball/" + __version__
